@@ -1,6 +1,5 @@
 /* main.c - main program and argument processing for cpio.
-   Copyright (C) 1990, 1991, 1992, 2001, 2003, 2004, 2005, 2006, 2007,
-   2009, 2010 Free Software Foundation, Inc.
+   Copyright (C) 1990-2019 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -36,6 +35,7 @@
 #endif
 
 #include <progname.h>
+#include <closeout.h>
 
 #include "filetypes.h"
 #include "cpiohdr.h"
@@ -57,7 +57,10 @@ enum cpio_options {
   FORCE_LOCAL_OPTION,            
   DEBUG_OPTION,                  
   BLOCK_SIZE_OPTION,             
-  TO_STDOUT_OPTION
+  TO_STDOUT_OPTION,
+  RENUMBER_INODES_OPTION,
+  IGNORE_DEVNO_OPTION,
+  DEVICE_INDEPENDENT_OPTION
 };
 
 const char *program_authors[] =
@@ -80,11 +83,13 @@ Examples:\n\
   # Copy files named in name-list to destination-directory\n\
   cpio -p destination-directory < name-list\n");
 
+static void usage (int status);
+
 /*  Print usage error message and exit with error.  */
 
 #define CHECK_USAGE(cond, opt, mode_opt) \
  if (cond) \
-   ERROR((PAXEXIT_FAILURE, 0, _("%s is meaningless with %s"), opt, mode_opt));
+   USAGE_ERROR ((0, 0, _("%s is meaningless with %s"), opt, mode_opt));
 
 static struct argp_option options[] = {
   /* ********** */
@@ -106,8 +111,9 @@ static struct argp_option options[] = {
   {NULL, 0, NULL, 0,
    N_("Operation modifiers valid in any mode:"), GRID },
 
-  {"file", 'F', N_("[[USER@]HOST:]FILE-NAME"), 0,
-   N_("Use this FILE-NAME instead of standard input or output. Optional USER and HOST specify the user and host names in case of a remote archive"), GRID+1 },
+  {"directory", 'D', N_("DIR"), 0,
+   N_("Change to directory DIR"), GRID+1 },
+  
   {"force-local", FORCE_LOCAL_OPTION, 0, 0,
    N_("Archive file is local, even if its name contains colons"), GRID+1 },
   {"format", 'H', N_("FORMAT"), 0,
@@ -117,21 +123,11 @@ static struct argp_option options[] = {
   {"block-size", BLOCK_SIZE_OPTION, N_("BLOCK-SIZE"), 0,
    N_("Set the I/O block size to BLOCK-SIZE * 512 bytes"), GRID+1 },
   {NULL, 'c', NULL, 0,
-   N_("Identical to \"-H newc\", use the new (SVR4) portable format.If you wish the old portable (ASCII) archive format, use \"-H odc\" instead."), GRID+1 },
+   N_("Identical to \"-H newc\", use the new (SVR4) portable format. If you wish the old portable (ASCII) archive format, use \"-H odc\" instead."), GRID+1 },
   {"dot", 'V', NULL, 0, 
    N_("Print a \".\" for each file processed"), GRID+1 },
   {"io-size", 'C', N_("NUMBER"), 0,
    N_("Set the I/O block size to the given NUMBER of bytes"), GRID+1 },
-  {"message", 'M', N_("STRING"), 0,
-   N_("Print STRING when the end of a volume of the backup media is reached"),
-   GRID+1 },
-  {"nonmatching", 'f', 0, 0,
-   N_("Only copy files that do not match any of the given patterns"), GRID+1 },
-  {"numeric-uid-gid", 'n', 0, 0,
-   N_("In the verbose table of contents listing, show numeric UID and GID"),
-   GRID+1 },
-  {"rsh-command", RSH_COMMAND_OPTION, N_("COMMAND"), 0,
-   N_("Use remote COMMAND instead of rsh"), GRID+1 },
   {"quiet", QUIET_OPTION, NULL, 0,
    N_("Do not print the number of blocks copied"), GRID+1 },
   {"verbose", 'v', NULL, 0,
@@ -142,12 +138,31 @@ static struct argp_option options[] = {
 #endif
   {"warning", 'W', N_("FLAG"), 0,
    N_("Control warning display. Currently FLAG is one of 'none', 'truncate', 'all'. Multiple options accumulate."), GRID+1 },
+  {"owner", 'R', N_("[USER][:.][GROUP]"), 0,
+   N_("Set the ownership of all files created to the specified USER and/or GROUP"), GRID+1 },
+#undef GRID
+
+#define GRID 110
+  {NULL, 0, NULL, 0,
+   N_("Operation modifiers valid in copy-in and copy-out modes"), GRID },
+  {"file", 'F', N_("[[USER@]HOST:]FILE-NAME"), 0,
+   N_("Use this FILE-NAME instead of standard input or output. Optional USER and HOST specify the user and host names in case of a remote archive"), GRID+1 },
+  {"message", 'M', N_("STRING"), 0,
+   N_("Print STRING when the end of a volume of the backup media is reached"),
+   GRID+1 },
+  {"rsh-command", RSH_COMMAND_OPTION, N_("COMMAND"), 0,
+   N_("Use COMMAND instead of rsh"), GRID+1 },
 #undef GRID
   
   /* ********** */
 #define GRID 200  
   {NULL, 0, NULL, 0,
    N_("Operation modifiers valid only in copy-in mode:"), GRID },
+  {"nonmatching", 'f', 0, 0,
+   N_("Only copy files that do not match any of the given patterns"), GRID+1 },
+  {"numeric-uid-gid", 'n', 0, 0,
+   N_("In the verbose table of contents listing, show numeric UID and GID"),
+   GRID+1 },
   {"pattern-file", 'E', N_("FILE"), 0,
    N_("Read additional patterns specifying filenames to extract or list from FILE"), 210},
   {"only-verify-crc", ONLY_VERIFY_CRC_OPTION, 0, 0,
@@ -165,6 +180,8 @@ static struct argp_option options[] = {
    GRID+1 },
   {"to-stdout", TO_STDOUT_OPTION, NULL, 0,
    N_("Extract files to standard output"), GRID+1 },
+  {NULL, 'I', N_("[[USER@]HOST:]FILE-NAME"), 0,
+   N_("Archive filename to use instead of standard input. Optional USER and HOST specify the user and host names in case of a remote archive"), GRID+1 },
 #undef GRID   
 
   /* ********** */
@@ -175,6 +192,13 @@ static struct argp_option options[] = {
    N_("Append to an existing archive."), GRID+1 },
   {NULL, 'O', N_("[[USER@]HOST:]FILE-NAME"), 0,
    N_("Archive filename to use instead of standard output. Optional USER and HOST specify the user and host names in case of a remote archive"), GRID+1 },
+  {"renumber-inodes", RENUMBER_INODES_OPTION, NULL, 0,
+   N_("Renumber inodes") },
+  {"ignore-devno", IGNORE_DEVNO_OPTION, NULL, 0,
+   N_("Don't store device numbers") },
+  {"device-independent", DEVICE_INDEPENDENT_OPTION, NULL, 0,
+   N_("Create device-independent (reproducible) archives") },
+  {"reproducible", 0, NULL, OPTION_ALIAS },
 #undef GRID
   
   /* ********** */
@@ -201,13 +225,9 @@ static struct argp_option options[] = {
   {NULL, 0, NULL, 0,
    N_("Operation modifiers valid in copy-out and copy-pass modes:"), GRID },
   {"null", '0', 0, 0,
-   N_("A list of filenames is terminated by a null character instead of a newline"), GRID+1 },
-  {NULL, 'I', N_("[[USER@]HOST:]FILE-NAME"), 0,
-   N_("Archive filename to use instead of standard input. Optional USER and HOST specify the user and host names in case of a remote archive"), GRID+1 },
+   N_("Filenames in the list are delimited by null characters instead of newlines"), GRID+1 },
   {"dereference", 'L', 0, 0,
    N_("Dereference  symbolic  links  (copy  the files that they point to instead of copying the links)."), GRID+1 },
-  {"owner", 'R', N_("[USER][:.][GROUP]"), 0,
-   N_("Set the ownership of all files created to the specified USER and/or GROUP"), GRID+1 },
   {"reset-access-time", 'a', NULL, 0,
    N_("Reset the access times of files after reading them"), GRID+1 },
 
@@ -300,14 +320,14 @@ parse_opt (int key, char *arg, struct argp_state *state)
 
     case BLOCK_SIZE_OPTION:		/* --block-size */
       io_block_size = atoi (arg);
-      if (io_block_size < 1)
-	error (2, 0, _("invalid block size"));
+      if (io_block_size < 1 || io_block_size > INT_MAX/512)
+	USAGE_ERROR ((0, 0, _("invalid block size")));
       io_block_size *= 512;
       break;
 
     case 'c':		/* Use the old portable ASCII format.  */
       if (archive_format != arf_unknown)
-	error (0, EXIT_FAILURE, _("Archive format multiply defined"));
+	USAGE_ERROR ((0, 0, _("Archive format multiply defined")));
 #define SVR4_COMPAT
 #ifdef SVR4_COMPAT
       archive_format = arf_newascii; /* -H newc.  */
@@ -319,13 +339,17 @@ parse_opt (int key, char *arg, struct argp_state *state)
     case 'C':		/* Block size.  */
       io_block_size = atoi (arg);
       if (io_block_size < 1)
-	error (2, 0, _("invalid block size"));
+	USAGE_ERROR ((0, 0, _("invalid block size")));
       break;
 
     case 'd':		/* Create directories where needed.  */
       create_dir_flag = true;
       break;
 
+    case 'D':
+      change_directory_option = arg;
+      break;
+      
     case 'f':		/* Only copy files not matching patterns.  */
       copy_matching_files = false;
       break;
@@ -340,7 +364,7 @@ parse_opt (int key, char *arg, struct argp_state *state)
 
     case 'H':		/* Header format name.  */
       if (archive_format != arf_unknown)
-	error (PAXEXIT_FAILURE, 0, _("Archive format multiply defined"));
+	USAGE_ERROR ((0, 0, _("Archive format multiply defined")));
       if (!strcasecmp (arg, "crc"))
 	archive_format = arf_crcascii;
       else if (!strcasecmp (arg, "newc"))
@@ -358,14 +382,14 @@ parse_opt (int key, char *arg, struct argp_state *state)
       else if (!strcasecmp (arg, "hpbin"))
 	archive_format = arf_hpbinary;
       else
-	error (2, 0, _("\
+	USAGE_ERROR ((0, 0, _("\
 invalid archive format `%s'; valid formats are:\n\
-crc newc odc bin ustar tar (all-caps also recognized)"), arg);
+crc newc odc bin ustar tar (all-caps also recognized)"), arg));
       break;
 	  
     case 'i':		/* Copy-in mode.  */
       if (copy_function != 0)
-	error (PAXEXIT_FAILURE, 0, _("Mode already defined"));
+	USAGE_ERROR ((0, 0, _("Mode already defined")));
       copy_function = process_copy_in;
       break;
 
@@ -373,11 +397,6 @@ crc newc odc bin ustar tar (all-caps also recognized)"), arg);
       input_archive_name = arg;
       break;
 
-    case 'k':		/* Handle corrupted archives.  We always handle
-			   corrupted archives, but recognize this
-			   option for compatability.  */
-      break;
-
     case 'l':		/* Link files when possible.  */
       link_flag = true;
       break;
@@ -408,14 +427,14 @@ crc newc odc bin ustar tar (all-caps also recognized)"), arg);
       
     case NO_PRESERVE_OWNER_OPTION:		/* --no-preserve-owner */
       if (set_owner_flag || set_group_flag)
-	error (PAXEXIT_FAILURE, 0, 
-               _("--no-preserve-owner cannot be used with --owner"));
+	USAGE_ERROR ((0, 0,  
+		      _("--no-preserve-owner cannot be used with --owner")));
       no_chown_flag = true;
       break;
 
     case 'o':		/* Copy-out mode.  */
       if (copy_function != 0)
-	error (PAXEXIT_FAILURE, 0, _("Mode already defined"));
+	USAGE_ERROR ((0, 0, _("Mode already defined")));
       copy_function = process_copy_out;
       break;
 
@@ -429,10 +448,22 @@ crc newc odc bin ustar tar (all-caps also recognized)"), arg);
 
     case 'p':		/* Copy-pass mode.  */
       if (copy_function != 0)
-	error (PAXEXIT_FAILURE, 0, _("Mode already defined"));
+	USAGE_ERROR ((0, 0, _("Mode already defined")));
       copy_function = process_copy_pass;
       break;
-	  
+
+    case IGNORE_DEVNO_OPTION:
+      ignore_devno_option = 1;
+      break;
+      
+    case RENUMBER_INODES_OPTION:
+      renumber_inodes_option = 1;
+      break;
+
+    case DEVICE_INDEPENDENT_OPTION:
+      ignore_devno_option = renumber_inodes_option = 1;
+      break;
+      
     case RSH_COMMAND_OPTION:
       rsh_command_option = arg;
       break;
@@ -451,15 +482,15 @@ crc newc odc bin ustar tar (all-caps also recognized)"), arg);
 
     case 'R':		/* Set the owner.  */
       if (no_chown_flag)
-	error (PAXEXIT_FAILURE, 0, 
-	       _("--owner cannot be used with --no-preserve-owner"));
+	USAGE_ERROR ((0, 0,  
+		      _("--owner cannot be used with --no-preserve-owner")));
       else
 	{
 	  char *e, *u, *g;
 	  
 	  e = parse_user_spec (arg, &set_owner, &set_group, &u, &g);
 	  if (e)
-	    error (PAXEXIT_FAILURE, 0, "%s: %s", arg, e);
+	    USAGE_ERROR ((0, 0, "%s: %s", arg, e));
 	  if (u)
 	    {
 	      free (u);
@@ -499,7 +530,7 @@ crc newc odc bin ustar tar (all-caps also recognized)"), arg);
 
     case 'W':
       if (warn_control (arg))
-	argp_error (state, _("Invalid value for --warning option: %s"), arg);
+	USAGE_ERROR ((0, 0, _("Invalid value for --warning option: %s"), arg));
       break;
       
     case SPARSE_OPTION:
@@ -536,14 +567,20 @@ static struct argp argp = {
   NULL
 };
 
+static void
+usage (int status)
+{
+  argp_help (&argp, stderr, ARGP_HELP_SEE, (char*) program_name);
+  close_stdout ();
+  exit (status);
+}
+
 /* Process the arguments.  Set all options and set up the copy pass
    directory or the copy in patterns.  */
 
 void
 process_args (int argc, char *argv[])
 {
-  void (*copy_in) ();		/* Work around for pcc bug.  */
-  void (*copy_out) ();
   int index;
 
   xstat = lstat;
@@ -558,36 +595,34 @@ process_args (int argc, char *argv[])
       if (table_flag)
 	copy_function = process_copy_in;
       else
-	error (PAXEXIT_FAILURE, 0, 
-	       _("You must specify one of -oipt options.\nTry `%s --help' or `%s --usage' for more information.\n"),
-	       program_name, program_name);
+	USAGE_ERROR ((0, 0,
+		      _("You must specify one of -oipt options.")));
     }
 
-  /* Work around for pcc bug.  */
-  copy_in = process_copy_in;
-  copy_out = process_copy_out;
-
-  if (copy_function == copy_in)
+  if (copy_function == process_copy_in)
     {
       archive_des = 0;
-      CHECK_USAGE(link_flag, "--link", "--extract");
-      CHECK_USAGE(reset_time_flag, "--reset", "--extract");
-      CHECK_USAGE(xstat != lstat, "--dereference", "--extract");
-      CHECK_USAGE(append_flag, "--append", "--extract");
-      CHECK_USAGE(output_archive_name, "-O", "--extract");
+      CHECK_USAGE (link_flag, "--link", "--extract");
+      CHECK_USAGE (reset_time_flag, "--reset", "--extract");
+      CHECK_USAGE (xstat != lstat, "--dereference", "--extract");
+      CHECK_USAGE (append_flag, "--append", "--extract");
+      CHECK_USAGE (output_archive_name, "-O", "--extract");
+      CHECK_USAGE (renumber_inodes_option, "--renumber-inodes", "--extract");
+      CHECK_USAGE (ignore_devno_option, "--ignore-devno", "--extract");
       if (to_stdout_option)
 	{
-	  CHECK_USAGE(create_dir_flag, "--make-directories", "--to-stdout");
-	  CHECK_USAGE(rename_flag, "--rename", "--to-stdout");
-	  CHECK_USAGE(no_chown_flag, "--no-preserve-owner", "--to-stdout");
-	  CHECK_USAGE(set_owner_flag||set_group_flag, "--owner", "--to-stdout");
-	  CHECK_USAGE(retain_time_flag, "--preserve-modification-time",
-		      "--to-stdout");
+	  CHECK_USAGE (create_dir_flag, "--make-directories", "--to-stdout");
+	  CHECK_USAGE (rename_flag, "--rename", "--to-stdout");
+	  CHECK_USAGE (no_chown_flag, "--no-preserve-owner", "--to-stdout");
+	  CHECK_USAGE (set_owner_flag||set_group_flag,
+		       "--owner", "--to-stdout");
+	  CHECK_USAGE (retain_time_flag, "--preserve-modification-time",
+		       "--to-stdout");
 	}
       
       if (archive_name && input_archive_name)
-	error (PAXEXIT_FAILURE, 0, 
-	       _("Both -I and -F are used in copy-in mode"));
+	USAGE_ERROR ((0, 0,  
+		      _("Both -I and -F are used in copy-in mode")));
 
       if (archive_format == arf_crcascii)
 	crc_i_flag = true;
@@ -596,72 +631,80 @@ process_args (int argc, char *argv[])
       if (input_archive_name)
 	archive_name = input_archive_name;
     }
-  else if (copy_function == copy_out)
+  else if (copy_function == process_copy_out)
     {
       if (index != argc)
-	error (PAXEXIT_FAILURE, 0, _("Too many arguments"));
+	USAGE_ERROR ((0, 0, _("Too many arguments")));
 
       archive_des = 1;
-      CHECK_USAGE(create_dir_flag, "--make-directories", "--create");
-      CHECK_USAGE(rename_flag, "--rename", "--create");
-      CHECK_USAGE(table_flag, "--list", "--create");
-      CHECK_USAGE(unconditional_flag, "--unconditional", "--create");
-      CHECK_USAGE(link_flag, "--link", "--create");
-      CHECK_USAGE(sparse_flag, "--sparse", "--create");
-      CHECK_USAGE(retain_time_flag, "--preserve-modification-time",
-		  "--create");
-      CHECK_USAGE(no_chown_flag, "--no-preserve-owner", "--create");
-      CHECK_USAGE(swap_bytes_flag, "--swap-bytes (--swap)", "--create");
-      CHECK_USAGE(swap_halfwords_flag, "--swap-halfwords (--swap)",
-		  "--create");
-      CHECK_USAGE(to_stdout_option, "--to-stdout", "--create");
+      CHECK_USAGE (create_dir_flag, "--make-directories", "--create");
+      CHECK_USAGE (rename_flag, "--rename", "--create");
+      CHECK_USAGE (table_flag, "--list", "--create");
+      CHECK_USAGE (unconditional_flag, "--unconditional", "--create");
+      CHECK_USAGE (link_flag, "--link", "--create");
+      CHECK_USAGE (sparse_flag, "--sparse", "--create");
+      CHECK_USAGE (retain_time_flag, "--preserve-modification-time",
+		   "--create");
+      CHECK_USAGE (no_chown_flag, "--no-preserve-owner", "--create");
+      CHECK_USAGE (swap_bytes_flag, "--swap-bytes (--swap)", "--create");
+      CHECK_USAGE (swap_halfwords_flag, "--swap-halfwords (--swap)",
+		   "--create");
+      CHECK_USAGE (to_stdout_option, "--to-stdout", "--create");
       
       if (append_flag && !(archive_name || output_archive_name))
-	error (PAXEXIT_FAILURE, 0, 
-	       _("--append is used but no archive file name is given (use -F or -O options)"));
+	USAGE_ERROR ((0, 0,  
+		      _("--append is used but no archive file name "
+			"is given (use -F or -O options)")));
 
-      CHECK_USAGE(rename_batch_file, "--rename-batch-file", "--create");
-      CHECK_USAGE(input_archive_name, "-I", "--create");
+      CHECK_USAGE (rename_batch_file, "--rename-batch-file", "--create");
+      CHECK_USAGE (input_archive_name, "-I", "--create");
       if (archive_name && output_archive_name)
-	error (PAXEXIT_FAILURE, 0, 
-	       _("Both -O and -F are used in copy-out mode"));
+	USAGE_ERROR ((0, 0,  
+		      _("Both -O and -F are used in copy-out mode")));
 
       if (archive_format == arf_unknown)
 	archive_format = arf_binary;
       if (output_archive_name)
 	archive_name = output_archive_name;
+
+      if (!arf_stores_inode_p (archive_format))
+	renumber_inodes_option = ignore_devno_option = 0;
     }
   else
     {
       /* Copy pass.  */
       if (index < argc - 1)
-	error (PAXEXIT_FAILURE, 0, _("Too many arguments"));
+	USAGE_ERROR ((0, 0, _("Too many arguments")));
       else if (index > argc - 1)
-	error (PAXEXIT_FAILURE, 0, _("Not enough arguments"));
+	USAGE_ERROR ((0, 0, _("Not enough arguments")));
 
       if (archive_format != arf_unknown)
-	error (PAXEXIT_FAILURE, 0, 
-	       _("Archive format is not specified in copy-pass mode (use --format option)"));
+	USAGE_ERROR ((0, 0,  
+		      _("Archive format is not specified in copy-pass mode "
+			"(use --format option)")));
 
-      CHECK_USAGE(swap_bytes_flag, "--swap-bytes (--swap)", "--pass-through");
-      CHECK_USAGE(swap_halfwords_flag, "--swap-halfwords (--swap)",
-		  "--pass-through");
-      CHECK_USAGE(table_flag, "--list", "--pass-through");
-      CHECK_USAGE(rename_flag, "--rename", "--pass-through");
-      CHECK_USAGE(append_flag, "--append", "--pass-through");
-      CHECK_USAGE(rename_batch_file, "--rename-batch-file", "--pass-through");
-      CHECK_USAGE(no_abs_paths_flag, "--no-absolute-pathnames",
-		  "--pass-through");
-      CHECK_USAGE(no_abs_paths_flag, "--absolute-pathnames",
-		  "--pass-through");
-      CHECK_USAGE(to_stdout_option, "--to-stdout", "--pass-through");
+      CHECK_USAGE (swap_bytes_flag, "--swap-bytes (--swap)", "--pass-through");
+      CHECK_USAGE (swap_halfwords_flag, "--swap-halfwords (--swap)",
+		   "--pass-through");
+      CHECK_USAGE (table_flag, "--list", "--pass-through");
+      CHECK_USAGE (rename_flag, "--rename", "--pass-through");
+      CHECK_USAGE (append_flag, "--append", "--pass-through");
+      CHECK_USAGE (rename_batch_file, "--rename-batch-file", "--pass-through");
+      CHECK_USAGE (no_abs_paths_flag, "--no-absolute-pathnames",
+		   "--pass-through");
+      CHECK_USAGE (no_abs_paths_flag, "--absolute-pathnames",
+		   "--pass-through");
+      CHECK_USAGE (to_stdout_option, "--to-stdout", "--pass-through");
+      CHECK_USAGE (renumber_inodes_option, "--renumber-inodes",
+		   "--pass-through");
+      CHECK_USAGE (ignore_devno_option, "--ignore-devno", "--pass-through");
       
       directory_name = argv[index];
     }
 
   if (archive_name)
     {
-      if (copy_function != copy_in && copy_function != copy_out)
+      if (copy_function != process_copy_in && copy_function != process_copy_out)
 	error (PAXEXIT_FAILURE, 0, 
 	       _("-F can be used only with --create or --extract"));
       archive_des = open_archive (archive_name);
